"""
We will from now on refer to macros as a set of standard codes that can be used
regardless of data sets or scenes in ParaView. ParaView's sense of macro is not
robust enough for us so that we will be referring to traditional macros in
ParaView as 'scripts' from here on.

*Scripts are used on specific sets of data whereas macros can be used on any
set of data.*

Macros, all of the ``pvmacros`` module, are standard codes that can be used
regardless of data sets or scenes in ParaView. These codes complete tedious or
recurring tasks either in ParaView's GUI or ParaView's batch processing
environment. We will use macros to complete everyday tasks like saving
screenshots of isometric views of a data scene or tedious tasks like making
many slices of a single data set along a line.
"""

from . import export
from . import vis
from .pipeline import *

__author__ = 'Bane Sullivan'
__license__ = 'BSD-3-Clause'
__copyright__ = '2018, Bane Sullivan'
__version__ = '2.0.1'


__displayname__ = 'ParaView Macros'



"""This module provides a complicated algorithm for making voxels out of regularly
gridded points. Considering that this algorithm is rather complex, we are keeping
it in its own module until we can simplify it, clean up the code, and make it
capable of handling non-uniformly gridded points
"""

__all__ = [
    'VoxelizePoints',
]

__displayname__ = 'Voxelize'

import numpy as np
import vtk
from vtk.numpy_interface import dataset_adapter as dsa
from vtk.util import numpy_support as nps

from .. import _helpers, interface
from ..base import FilterBase
from ..version import check_numpy
from .xyz import RotationTool

###############################################################################


class VoxelizePoints(FilterBase):
    """This makes a ``vtkUnstructuredGrid`` of scattered points given voxel
    sizes as input arrays. This assumes that the data is at least 2-Dimensional
    on the XY Plane.
    """
    __displayname__ = 'Voxelize Points'
    __category__ = 'filter'
    def __init__(self, **kwargs):
        FilterBase.__init__(self,
            nInputPorts=1, inputType='vtkPointSet',
            nOutputPorts=1, outputType='vtkUnstructuredGrid')
        self.__dx = kwargs.get('dx', None)
        self.__dy = kwargs.get('dy', None)
        self.__dz = kwargs.get('dz', None)
        self.__estimateGrid = kwargs.get('estimate', True)
        self.__safe = kwargs.get('safe', 10.0)

        # Not controlled by user:
        self.__angle = 0.0



    def add_field_data(self, grid):
        """An internal helper to add the recovered information as field data
        """
        # Add angle
        a = vtk.vtkDoubleArray()
        a.SetName('Recovered Angle (Deg.)')
        a.SetNumberOfValues(1)
        a.SetValue(0, np.rad2deg(self.__angle))
        grid.GetFieldData().AddArray(a)
        # Add cell sizes
        s = vtk.vtkDoubleArray()
        s.SetName('Recovered Cell Sizes')
        s.SetNumberOfComponents(3)
        s.InsertNextTuple3(self.__dx, self.__dy, self.__dz)
        grid.GetFieldData().AddArray(s)
        return grid

    @staticmethod
    def add_cell_data(grid, arr, name):
        """Add a NumPy array as cell data to the given grid input
        """
        c = interface.convert_array(arr, name=name)
        grid.GetCellData().AddArray(c)
        return grid


    def estimate_uniform_spacing(self, x, y, z):
        """This assumes that the input points make up some sort of uniformly
        spaced grid on at least an XY plane.
        """
        # TODO: implement ability to rotate around Z axis (think PoroTomo vs UTM)
        # TODO: implement way to estimate rotation
        if not (len(x) == len(y) == len(z)):
            raise AssertionError('Must have same number of coordinates for all components.')
        num = len(x)
        if num == 1:
            # Only one point.. use safe
            return x, y, z, self.__safe, self.__safe, self.__safe, 0.0

        r = RotationTool()
        xr, yr, zr, dx, dy, angle = r.estimate_and_rotate(x, y, z)
        self.__angle = angle
        uz = np.diff(np.unique(z))
        if len(uz) > 0: dz = np.average(uz)
        else: dz = self.__safe
        self.__dx = dx
        self.__dy = dy
        self.__dz = dz
        return xr, yr, zr


    def points_to_grid(self, xo,yo,zo, dx,dy,dz, grid=None):
        """Convert XYZ points to a ``vtkUnstructuredGrid``.
        """
        if not check_numpy(alert='warn'):
            return grid
        if grid is None:
            grid = vtk.vtkUnstructuredGrid()

        # TODO: Check dtypes on all arrays. Need to be floats

        if self.__estimateGrid:
            x,y,z = self.estimate_uniform_spacing(xo, yo, zo)
        else:
            x,y,z = xo, yo, zo

        dx,dy,dz = self.__dx, self.__dy, self.__dz
        if isinstance(dx, np.ndarray) and len(dx) != len(x):
            raise _helpers.PVGeoError('X-Cell spacings are not properly defined for all points.')
        if isinstance(dy, np.ndarray) and len(dy) != len(y):
            raise _helpers.PVGeoError('X-Cell spacings are not properly defined for all points.')
        if isinstance(dz, np.ndarray) and len(dz) != len(z):
            raise _helpers.PVGeoError('X-Cell spacings are not properly defined for all points.')

        numCells = len(x)

        # Generate cell nodes for all points in data set
        #- Bottom
        c_n1 = np.stack( ((x - dx/2) , (y - dy/2), (z - dz/2) ), axis=1)
        c_n2 = np.stack(( (x + dx/2) , (y - dy/2), (z - dz/2) ), axis=1)
        c_n3 = np.stack(( (x - dx/2) , (y + dy/2), (z - dz/2) ), axis=1)
        c_n4 = np.stack(( (x + dx/2) , (y + dy/2), (z - dz/2) ), axis=1)
        #- Top
        c_n5 = np.stack(( (x - dx/2) , (y - dy/2), (z + dz/2) ), axis=1)
        c_n6 = np.stack(( (x + dx/2) , (y - dy/2), (z + dz/2) ), axis=1)
        c_n7 = np.stack(( (x - dx/2) , (y + dy/2), (z + dz/2) ), axis=1)
        c_n8 = np.stack(( (x + dx/2) , (y + dy/2), (z + dz/2) ), axis=1)

        #- Concatenate
        all_nodes = np.concatenate((
            c_n1,
            c_n2,
            c_n3,
            c_n4,
            c_n5,
            c_n6,
            c_n7,
            c_n8), axis=0)

        # Search for unique nodes and use the min cell size as the tolerance
        TOLERANCE = np.min([dx, dy]) / 2.0
        # Round XY plane by the tolerance
        txy = np.around(all_nodes[:,0:2]/TOLERANCE)
        all_nodes[:,0:2] = txy
        unique_nodes, ind_nodes = np.unique(all_nodes, return_inverse=True, axis=0)
        unique_nodes[:,0:2] *= TOLERANCE
        numPts = len(unique_nodes)

        # Make the cells
        pts = vtk.vtkPoints()
        cells = vtk.vtkCellArray()

        # insert unique nodes as points
        if self.__estimateGrid:
            unique_nodes[:,0:2] = RotationTool.rotate(unique_nodes[:,0:2], -self.__angle)
            self.add_field_data(grid)

        # Add unique nodes as points in output
        pts.SetData(interface.convert_array(unique_nodes))

        # Add cell vertices
        j = np.multiply(np.tile(np.arange(0, 8, 1), numCells), numCells)
        arridx = np.add(j, np.repeat(np.arange(0, numCells, 1, dtype=int), 8))

        ids = ind_nodes[arridx].reshape((numCells, 8))
        cellsMat = np.concatenate((np.ones((ids.shape[0], 1), dtype=np.int64)*ids.shape[1], ids), axis=1).ravel()

        cells = vtk.vtkCellArray()
        cells.SetNumberOfCells(numCells)
        cells.SetCells(numCells, nps.numpy_to_vtk(cellsMat, deep=True, array_type=vtk.VTK_ID_TYPE))

        # Set the output
        grid.SetPoints(pts)
        grid.SetCells(vtk.VTK_VOXEL, cells)
        return grid

    @staticmethod
    def _copy_arrays(pdi, pdo):
        """internal helper to copy arrays from point data to cell data in the voxels.
        """
        for i in range(pdi.GetPointData().GetNumberOfArrays()):
            arr = pdi.GetPointData().GetArray(i)
            _helpers.add_array(pdo, 1, arr) # adds to CELL data
        return pdo

    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline to generate output
        """
        # Get input/output of Proxy
        pdi = self.GetInputData(inInfo, 0, 0)
        pdo = self.GetOutputData(outInfo, 0)
        # Perfrom task
        wpdi = dsa.WrapDataObject(pdi)
        pts = wpdi.Points
        x, y, z = pts[:,0], pts[:,1], pts[:,2]
        self.points_to_grid(x, y, z,
            self.__dx, self.__dy, self.__dz, grid=pdo)
        # Now append data to grid
        self._copy_arrays(pdi, pdo)
        return 1


    #### Seters and Geters ####


    def set_safe_size(self, safe):
        """A voxel size to use if a spacing cannot be determined for an axis
        """
        if self.__safe != safe:
            self.__safe = safe
            self.Modified()

    def set_delta_x(self, dx):
        """Set the X cells spacing

        Args:
            dx (float or np.array(floats)): the spacing(s) for the cells in
                the X-direction
        """
        self.__dx = dx
        self.Modified()

    def set_delta_y(self, dy):
        """Set the Y cells spacing

        Args:
            dy (float or np.array(floats)): the spacing(s) for the cells in
                the Y-direction
        """
        self.__dy = dy
        self.Modified()

    def set_delta_z(self, dz):
        """Set the Z cells spacing

        Args:
            dz (float or np.array(floats)): the spacing(s) for the cells in
                the Z-direction
        """
        self.__dz = dz
        self.set_safe_size(np.min(dz))
        self.Modified()

    def set_deltas(self, dx, dy, dz):
        """Set the cell spacings for each axial direction

        Args:
            dx (float or np.array(floats)): the spacing(s) for the cells in
                the X-direction
            dy (float or np.array(floats)): the spacing(s) for the cells in
                the Y-direction
            dz (float or np.array(floats)): the spacing(s) for the cells in
                the Z-direction
        """
        self.set_delta_x(dx)
        self.set_delta_y(dy)
        self.set_delta_z(dz)

    def set_estimate_grid(self, flag):
        """Set a flag on whether or not to estimate the grid spacing/rotation
        """
        if self.__estimateGrid != flag:
            self.__estimateGrid = flag
            self.Modified()


    def get_recovered_angle(self, degrees=True):
        """Returns the recovered angle if set to recover the input grid. If the
        input points are rotated, then this angle will reflect a close
        approximation of that rotation.

        Args:
            degrees (bool): A flag on to return decimal degrees or radians.
        """
        if degrees: return np.rad2deg(self.__angle)
        return self.__angle

    def get_spacing(self):
        """Get the cell spacings"""
        return (self.__dx, self.__dy, self.__dz)




###############################################################################



__all__ = [
    'ExtractTopography',
]

__displayname__ = 'Subsetting'

import numpy as np
import vtk
from vtk.numpy_interface import dataset_adapter as dsa

from .. import _helpers, interface
from ..base import FilterBase

# NOTE: internal import - from scipy.spatial import cKDTree



###############################################################################

class ExtractTopography(FilterBase):
    """This filter takes two inputs: any mesh dataset and a set of points for
    a topography source. This will add a boolean data array to the cell data of
    the input grid on whether that cell should be active (under topographic
    layer). A user can also choose to directly extract the data rather than
    appending a boolean scalar array via the ``remove`` argument.

    Args:
        op (str, int, or callable): The operation as a string key, int
            index, or callable method

        tolerance (float): buffer around the topography surface to include as
            part of the decision boundary

        offset (float): static value to shift the reference topography surface

        ivert (bool): optional to invert the extraction.

        remove (bool): Optional parameter to apply a thresholding filter and
            return a ``vtkUnstructuredGrid`` object with only the extracted
            cells. The ``remove`` option is only available in Python
            environments (not available in ParaView). The ``remove`` flag must
            be set at the time of instantiation of this algorithm.
            This does not actually update the algorithm's output data object
            but applies a `PyVista` threshold filter to pass a new data object
            after calling ``apply``.


    Note:
        This currenlty ignores time varying inputs. We can implement time
        variance but need to think about how we would like to do that. Should
        the topography surface be static and the volumetric data have time
        variance?

    """
    __displayname__ = 'Extract Topography'
    __category__ = 'filter'
    def __init__(self, op='underneath', tolerance=0.001, offset=0.0,
                 invert=False, remove=False):
        FilterBase.__init__(self,
            nInputPorts=2, inputType='vtkDataSet',
            nOutputPorts=1)
        self._tolerance = tolerance
        self._offset = offset
        self._invert = invert
        self._remove = remove
        self._operation = self._underneath
        self.set_operation(op)

    # CRITICAL for multiple input ports
    def FillInputPortInformation(self, port, info):
        """This simply makes sure the user selects the correct inputs
        """
        typ = 'vtkDataSet'
        if port == 1:
            typ = 'vtkPointSet' # Make sure topography is some sort of point set
        info.Set(self.INPUT_REQUIRED_DATA_TYPE(), typ)
        return 1

    # THIS IS CRUCIAL to preserve data type through filter
    def RequestDataObject(self, request, inInfo, outInfo):
        """Constructs the output data object based on the input data object
        """
        self.OutputType = self.GetInputData(inInfo, 0, 0).GetClassName()
        self.FillOutputPortInformation(0, outInfo.GetInformationObject(0))
        return 1


    #### Extraction Methods ####

    @staticmethod
    def _query(topo_points, data_points):
        """Querrys the data points for their closest point on the topography
        surface"""
        try:
            # sklearn's KDTree is faster: use it if available
            from sklearn.neighbors import KDTree as Tree
        except ImportError:
            from scipy.spatial import cKDTree  as Tree
        tree = Tree(topo_points)
        i = tree.query(data_points)[1].ravel()
        return topo_points[i]

    @staticmethod
    def _underneath(topo_points, data_points, tolerance):
        """Extract cells underneath the topography surface"""
        comp = ExtractTopography._query(topo_points, data_points)
        return np.array(data_points[:,2] < (comp[:,2] - tolerance), dtype=int)

    @staticmethod
    def _intersection(topo_points, data_points, tolerance):
        """Extract cells intersecting the topography surface"""
        comp = ExtractTopography._query(topo_points, data_points)
        return np.array(np.abs((data_points[:,2] - comp[:,2])) < tolerance, dtype=int)

    @staticmethod
    def get_operations():
        """Returns the extraction operation methods as callable objects in a
        dictionary
        """
        ops = dict(
            underneath=ExtractTopography._underneath,
            intersection=ExtractTopography._intersection,
        )
        return ops

    @staticmethod
    def get_operation_names():
        """Gets a list of the extraction operation keys

        Return:
            list(str): the keys for getting the operations
        """
        ops = ExtractTopography.get_operations()
        return list(ops.keys())

    @staticmethod
    def get_operation(idx):
        """Gets a extraction operation based on an index in the keys

        Return:
            callable: the operation method
        """
        if isinstance(idx, str):
            return ExtractTopography.get_operations()[idx]
        n = ExtractTopography.get_operation_names()[idx]
        return ExtractTopography.get_operations()[n]


    #### Pipeline Methods ####

    def RequestData(self, request, inInfo, outInfo):
        """Used by pipeline to generate output"""
        # Get input/output of Proxy
        igrid = self.GetInputData(inInfo, 0, 0) # Port 0: grid
        topo = self.GetInputData(inInfo, 1, 0) # Port 1: topography
        grid = self.GetOutputData(outInfo, 0)
        grid.DeepCopy(igrid)

        # Perfrom task
        ncells = igrid.GetNumberOfCells()
        active = np.zeros((ncells), dtype=int)
        # Now iterate through the cells in the grid and test if they are beneath the topography
        wtopo = dsa.WrapDataObject(topo)
        topo_points = np.array(wtopo.Points) # mak sure we do not edit the input
        #  shift the topography points for the tree
        topo_points[:,2] = topo_points[:,2] + self._offset

        filt = vtk.vtkCellCenters()
        filt.SetInputDataObject(igrid)
        filt.Update()
        data_points = dsa.WrapDataObject(filt.GetOutput(0)).Points

        active = self._operation(topo_points, data_points, self._tolerance)

        if self._invert:
            # NOTE: assumes the given operation produces zeros and ones only
            active = 1 - active

        # Now add cell data to output
        active = interface.convert_array(active, name='Extracted')
        grid.GetCellData().AddArray(active)
        return 1

    def apply(self, data, points):
        """Run the algorithm on the input data using the topography points"""
        self.SetInputDataObject(0, data)
        self.SetInputDataObject(1, points)
        self.Update()
        output = interface.wrap_pyvista(self.GetOutput())
        if self._remove:
            # NOTE: Assumes the given operation produces zeros and ones only
            #       Also, this does not update the algorithm's output.
            #       This only sends a new thresholded dataset to the user.
            return output.threshold(0.5, scalars='Extracted')
        return output

    #### Setters/Getters ####

    def set_tolerance(self, tol):
        """Set the tolerance threshold for the querry"""
        if self._tolerance != tol:
            self._tolerance = tol
            self.Modified()

    def get_tolerance(self):
        """Get the tolerance threshold for the querry"""
        return self._tolerance

    def set_offset(self, offset):
        """Sets how far off (in Z dir) to slice the data"""
        if self._offset != offset:
            self._offset = offset
            self.Modified()

    def set_invert(self, flag):
        """Sets the boolean flag on whether to invert the extraction."""
        if self._invert != flag:
            self._invert = flag
            self.Modified()

    def set_operation(self, op):
        """Set the type of extraction to perform.

        Args:
            op (str, int, or callable): The operation as a string key, int
                index, or callable method

        Note:
            This can accept a callable method to set a custom operation as long
            as its signature is ``<callable>(self, topo_points, data_points)`` and it
            strictly produces an integer array of zeros and ones.
        """
        if isinstance(op, str):
            op = ExtractTopography.get_operations()[op]
        elif isinstance(op, int):
            op = ExtractTopography.get_operation(op)
        if self._operation != op:
            self._operation = op
            self.Modified()



paraview_plugin_version = '2.0.1'
# This is module to import. It provides VTKPythonAlgorithmBase, the base class
# for all python-based vtkAlgorithm subclasses in VTK and decorators used to
# 'register' the algorithm with ParaView along with information about UI.
from paraview.util.vtkAlgorithm import *

# Helpers:
from PVGeo import _helpers
# Classes to Decorate
from PVGeo.ubc import *

discretize_available = False
try:
    with _helpers.HiddenPrints():
        import discretize
except ImportError:
    pass
else:
    discretize_available = True

#### GLOBAL VARIABLES ####
MENU_CAT = 'PVGeo: UBC Mesh Tools'



@smproxy.reader(name="PVGeoTensorMeshReader",
       label='PVGeo: %s'%TensorMeshReader.__displayname__,
       extensions=TensorMeshReader.extensions,
       file_description=TensorMeshReader.description)
@smhint.xml('''<RepresentationType view="RenderView" type="Surface With Edges" />''')
class PVGeoTensorMeshReader(TensorMeshReader):
    def __init__(self):
        TensorMeshReader.__init__(self)


    #### Seters and Geters ####

    @smproperty.xml('''
        <StringVectorProperty
            panel_visibility="advanced"
            name="MeshFile"
            label="File Name Mesh"
            command="set_mesh_filename"
            animateable="1"
            clean_command="clear_mesh"
            number_of_elements="1">
            <FileListDomain name="meshfile"/>
            <Documentation>This is the mesh file for a 2D or 3D UBC Mesh grid. This plugin only allows ONE mesh to be defined.</Documentation>
        </StringVectorProperty>''')
    def set_mesh_filename(self, filename):
        TensorMeshReader.set_mesh_filename(self, filename)

    @smproperty.xml('''
        <StringVectorProperty
          panel_visibility="default"
          name="ModelFiles"
          label="File Name(s) Model"
          command="add_model_file_name"
          animateable="1"
          repeat_command="1"
          clean_command="clear_models"
          number_of_elements="1">
          <FileListDomain name="model_files"/>
          <Documentation>This is for a single sets of model files to append to the mesh as data time varying attributes. You can chose as many files as you would like for this for the given attribute.</Documentation>
        </StringVectorProperty>''')
    def add_model_file_name(self, filename):
        """Use to set the file names for the reader. Handles singlt string or list of strings."""
        TensorMeshReader.add_model_file_name(self, filename)


    @smproperty.doublevector(name="TimeDelta", default_values=1.0, panel_visibility="advanced")
    def set_time_delta(self, dt):
        TensorMeshReader.set_time_delta(self, dt)

    @smproperty.doublevector(name="TimestepValues", information_only="1", si_class="vtkSITimeStepsProperty")
    def get_time_step_values(self):
        """This is critical for registering the timesteps"""
        return TensorMeshReader.get_time_step_values(self)

    @smproperty.xml(_helpers.get_property_xml(name='Use file as data name', command='set_use_filename', default_values=True, help='A boolean to override the DataName and use model file name as data name.',
    panel_visibility="advanced"))
    def set_use_filename(self, flag):
        TensorMeshReader.set_use_filename(self, flag)

    @smproperty.stringvector(name='DataName', default_values='Data', panel_visibility="advanced")
    def set_data_name(self, name):
        TensorMeshReader.set_data_name(self, name)

@smproxy.filter(name="PVGeoTensorMeshAppender",
       label=TensorMeshAppender.__displayname__)
@smhint.xml('''<ShowInMenu category="%s"/>
    <RepresentationType view="RenderView" type="Surface" />''' % MENU_CAT)
@smproperty.input(name="Input", port_index=0)
@smdomain.datatype(dataTypes=["vtkRectilinearGrid"], composite_data_supported=False)
class PVGeoTensorMeshAppender(TensorMeshAppender):
    """This assumes the input vtkRectilinearGrid has already handled the timesteps"""
    def __init__(self):
        TensorMeshAppender.__init__(self)

    @smproperty.xml('''
        <StringVectorProperty
          panel_visibility="default"
          name="ModelFiles"
          label="File Name(s) Model"
          command="add_model_file_name"
          animateable="1"
          repeat_command="1"
          clean_command="clear_models"
          number_of_elements="1">
          <FileListDomain name="model_files"/>
          <Documentation>This is for a single sets of model files to append to the mesh as data time varying attributes. You can chose as many files as you would like for this for the given attribute.</Documentation>
        </StringVectorProperty>''')
    def add_model_file_name(self, filename):
        """Use to set the file names for the reader. Handles single string or list of strings."""
        TensorMeshAppender.add_model_file_name(self, filename)

    @smproperty.xml(_helpers.get_property_xml(name='Use file as data name', command='set_use_filename', default_values=True, help='A boolean to override the DataName and use model file name as data name.',
    panel_visibility="advanced"))
    def set_use_filename(self, flag):
        TensorMeshAppender.set_use_filename(self, flag)

    @smproperty.stringvector(name='DataName', default_values='Appended Data', panel_visibility="advanced")
    def set_data_name(self, name):
        TensorMeshAppender.set_data_name(self, name)

    @smproperty.doublevector(name="TimestepValues", information_only="1", si_class="vtkSITimeStepsProperty")
    def get_time_step_values(self):
        """This is critical for registering the timesteps"""
        return TensorMeshAppender.get_time_step_values(self)



@smproxy.filter(name="PVGeoTopoMeshAppender",
       label=TopoMeshAppender.__displayname__)
@smhint.xml('''<ShowInMenu category="%s"/>
    <RepresentationType view="RenderView" type="Surface" />''' % MENU_CAT)
@smproperty.input(name="Input", port_index=0)
@smdomain.datatype(dataTypes=["vtkRectilinearGrid"], composite_data_supported=False)
class PVGeoTopoMeshAppender(TopoMeshAppender):
    """This assumes the input vtkRectilinearGrid has already handled the timesteps"""
    def __init__(self):
        TopoMeshAppender.__init__(self)

    @smproperty.xml('''
        <StringVectorProperty
            panel_visibility="advanced"
            name="TopoFile"
            label="File Name Topo"
            command="set_topo_filename"
            animateable="1"
            clean_command="clear_topo_file"
            number_of_elements="1">
            <FileListDomain name="topofile"/>
            <Documentation>This plugin only allows ONE topo file to be defined.</Documentation>
        </StringVectorProperty>''')
    def set_topo_filename(self, filename):
        TopoMeshAppender.set_topo_filename(self, filename)




#------------------------------------------------------------------------------
# Read OcTree Mesh
#------------------------------------------------------------------------------

if discretize_available:
    @smproxy.reader(name="PVGeoUBCOcTreeMeshReader",
           label='PVGeo: %s'%OcTreeReader.__displayname__,
           extensions=OcTreeReader.extensions,
           file_description=OcTreeReader.description)
    @smhint.xml('''<RepresentationType view="RenderView" type="Surface With Edges" />''')
    class PVGeoUBCOcTreeMeshReader(OcTreeReader):
        def __init__(self):
            OcTreeReader.__init__(self)


        #### Seters and Geters ####

        @smproperty.xml('''
            <StringVectorProperty
                panel_visibility="advanced"
                name="MeshFile"
                label="File Name Mesh"
                command="set_mesh_filename"
                animateable="1"
                clean_command="clear_mesh"
                number_of_elements="1">
                <FileListDomain name="meshfile"/>
                <Documentation>This is the mesh file for a OcTree Mesh grid. This plugin only allows ONE mesh to be defined.</Documentation>
            </StringVectorProperty>''')
        def set_mesh_filename(self, filename):
            OcTreeReader.set_mesh_filename(self, filename)

        @smproperty.xml('''
            <StringVectorProperty
              panel_visibility="default"
              name="ModelFiles"
              label="File Name(s) Model"
              command="add_model_file_name"
              animateable="1"
              repeat_command="1"
              clean_command="clear_models"
              number_of_elements="1">
              <FileListDomain name="model_files"/>
              <Documentation>This is for a single sets of model files to append to the mesh as data time varying attributes. You can chose as many files as you would like for this for the given attribute.</Documentation>
            </StringVectorProperty>''')
        def add_model_file_name(self, filename):
            """Use to set the file names for the reader. Handles singlt string or list of strings."""
            OcTreeReader.add_model_file_name(self, filename)


        @smproperty.doublevector(name="TimeDelta", default_values=1.0, panel_visibility="advanced")
        def set_time_delta(self, dt):
            OcTreeReader.set_time_delta(self, dt)

        @smproperty.doublevector(name="TimestepValues", information_only="1", si_class="vtkSITimeStepsProperty")
        def get_time_step_values(self):
            """This is critical for registering the timesteps"""
            return OcTreeReader.get_time_step_values(self)

        @smproperty.xml(_helpers.get_property_xml(name='Use file as data name', command='set_use_filename', default_values=True, help='A boolean to override the DataName and use model file name as data name.',
        panel_visibility="advanced"))
        def set_use_filename(self, flag):
            OcTreeReader.set_use_filename(self, flag)

        @smproperty.stringvector(name='DataName', default_values='Data', panel_visibility="advanced")
        def set_data_name(self, name):
            OcTreeReader.set_data_name(self, name)



    @smproxy.filter(name="PVGeoOcTreeAppender",
           label=OcTreeAppender.__displayname__)
    @smhint.xml('''<ShowInMenu category="%s"/>
        <RepresentationType view="RenderView" type="Surface With Edges" />''' % MENU_CAT)
    @smproperty.input(name="Input", port_index=0)
    @smdomain.datatype(dataTypes=["vtkUnstructuredGrid"], composite_data_supported=False)
    class PVGeoOcTreeAppender(OcTreeAppender):
        """This assumes the input vtkUnstructuredGrid has already handled the timesteps"""
        def __init__(self):
            OcTreeAppender.__init__(self)

        @smproperty.xml('''
            <StringVectorProperty
              panel_visibility="default"
              name="ModelFiles"
              label="File Name(s) Model"
              command="add_model_file_name"
              animateable="1"
              repeat_command="1"
              clean_command="clear_models"
              number_of_elements="1">
              <FileListDomain name="model_files"/>
              <Documentation>This is for a single sets of model files to append to the mesh as data time varying attributes. You can chose as many files as you would like for this for the given attribute.</Documentation>
            </StringVectorProperty>''')
        def add_model_file_name(self, filename):
            """Use to set the file names for the reader. Handles single string or list of strings."""
            OcTreeAppender.add_model_file_name(self, filename)

        @smproperty.xml(_helpers.get_property_xml(name='Use file as data name', command='set_use_filename', default_values=True, help='A boolean to override the DataName and use model file name as data name.',
        panel_visibility="advanced"))
        def set_use_filename(self, flag):
            OcTreeAppender.set_use_filename(self, flag)

        @smproperty.stringvector(name='DataName', default_values='Appended Data', panel_visibility="advanced")
        def set_data_name(self, name):
            OcTreeAppender.set_data_name(self, name)


#------------------------------------------------------------------------------
# Write Tensor Mesh
#------------------------------------------------------------------------------


@smproxy.writer(extensions="msh", file_description="UBC Tensor Mesh", support_reload=False)
@smproperty.input(name="Input", port_index=0)
@smdomain.datatype(dataTypes=["vtkRectilinearGrid"], composite_data_supported=True)
class PVGeoWriteRectilinearGridToUBC(WriteRectilinearGridToUBC):
    def __init__(self):
        WriteRectilinearGridToUBC.__init__(self)

    @smproperty.stringvector(name="FileName", panel_visibility="never")
    @smdomain.filelist()
    def SetFileName(self, filename):
        """Specify filename for the file to write."""
        WriteRectilinearGridToUBC.SetFileName(self, filename)

    @smproperty.stringvector(name="Format", default_values='%.9e')
    def set_format(self, fmt):
        """Use to set the ASCII format for the writer default is ``'%.9e'``"""
        WriteRectilinearGridToUBC.set_format(self, fmt)


@smproxy.writer(extensions="msh", file_description="UBC Tensor Mesh", support_reload=False)
@smproperty.input(name="Input", port_index=0)
@smdomain.datatype(dataTypes=["vtkImageData"], composite_data_supported=True)
class PVGeoWriteImageDataToUBC(WriteImageDataToUBC):
    def __init__(self):
        WriteImageDataToUBC.__init__(self)

    @smproperty.stringvector(name="FileName", panel_visibility="never")
    @smdomain.filelist()
    def SetFileName(self, filename):
        """Specify filename for the file to write."""
        WriteImageDataToUBC.SetFileName(self, filename)

    @smproperty.stringvector(name="Format", default_values='%.9e')
    def set_format(self, fmt):
        """Use to set the ASCII format for the writer default is ``'%.9e'``"""
        WriteImageDataToUBC.set_format(self, fmt)


###############################################################################


@smproxy.reader(name="PVGeoTopoReader",
       label='PVGeo: %s'%TopoReader.__displayname__,
       extensions=TopoReader.extensions,
       file_description=TopoReader.description)
class PVGeoTopoReader(TopoReader):
    def __init__(self):
        TopoReader.__init__(self)

    #### Seters and Geters ####
    @smproperty.xml(_helpers.get_file_reader_xml(TopoReader.extensions, reader_description=TopoReader.description))
    def AddFileName(self, filename):
        TopoReader.AddFileName(self, filename)

    @smproperty.doublevector(name="TimeDelta", default_values=1.0, panel_visibility="advanced")
    def set_time_delta(self, dt):
        TopoReader.set_time_delta(self, dt)

    @smproperty.doublevector(name="TimestepValues", information_only="1", si_class="vtkSITimeStepsProperty")
    def get_time_step_values(self):
        """This is critical for registering the timesteps"""
        return TopoReader.get_time_step_values(self)

    @smproperty.intvector(name="SkipRows", default_values=0, panel_visibility="advanced")
    def set_skip_rows(self, skip):
        TopoReader.set_skip_rows(self, skip)

    @smproperty.stringvector(name="Comments", default_values="!", panel_visibility="advanced")
    def set_comments(self, identifier):
        TopoReader.set_comments(self, identifier)



###############################################################################


@smproxy.reader(name="PVGeoGravObsReader",
       label='PVGeo: %s'%GravObsReader.__displayname__,
       extensions=GravObsReader.extensions,
       file_description=GravObsReader.description)
class PVGeoGravObsReader(GravObsReader):
    def __init__(self):
        GravObsReader.__init__(self)

    #### Seters and Geters ####
    @smproperty.xml(_helpers.get_file_reader_xml(GravObsReader.extensions, reader_description=GravObsReader.description))
    def AddFileName(self, filename):
        GravObsReader.AddFileName(self, filename)

    @smproperty.doublevector(name="TimeDelta", default_values=1.0, panel_visibility="advanced")
    def set_time_delta(self, dt):
        GravObsReader.set_time_delta(self, dt)

    @smproperty.doublevector(name="TimestepValues", information_only="1", si_class="vtkSITimeStepsProperty")
    def get_time_step_values(self):
        """This is critical for registering the timesteps"""
        return GravObsReader.get_time_step_values(self)

    @smproperty.intvector(name="SkipRows", default_values=0, panel_visibility="advanced")
    def set_skip_rows(self, skip):
        GravObsReader.set_skip_rows(self, skip)

    @smproperty.stringvector(name="Comments", default_values="!", panel_visibility="advanced")
    def set_comments(self, identifier):
        GravObsReader.set_comments(self, identifier)


###############################################################################


@smproxy.reader(name="PVGeoGravGradReader",
       label='PVGeo: %s'%GravGradReader.__displayname__,
       extensions=GravGradReader.extensions,
       file_description=GravGradReader.description)
class PVGeoGravGradReader(GravGradReader):
    def __init__(self):
        GravGradReader.__init__(self)

    #### Seters and Geters ####
    @smproperty.xml(_helpers.get_file_reader_xml(GravGradReader.extensions, reader_description=GravGradReader.description))
    def AddFileName(self, filename):
        GravGradReader.AddFileName(self, filename)

    @smproperty.doublevector(name="TimeDelta", default_values=1.0, panel_visibility="advanced")
    def set_time_delta(self, dt):
        GravGradReader.set_time_delta(self, dt)

    @smproperty.doublevector(name="TimestepValues", information_only="1", si_class="vtkSITimeStepsProperty")
    def get_time_step_values(self):
        """This is critical for registering the timesteps"""
        return GravGradReader.get_time_step_values(self)

    @smproperty.intvector(name="SkipRows", default_values=0, panel_visibility="advanced")
    def set_skip_rows(self, skip):
        GravGradReader.set_skip_rows(self, skip)

    @smproperty.stringvector(name="Comments", default_values="!", panel_visibility="advanced")
    def set_comments(self, identifier):
        GravGradReader.set_comments(self, identifier)



###############################################################################


@smproxy.reader(name="PVGeoMagObsReader",
       label='PVGeo: %s'%MagObsReader.__displayname__,
       extensions=MagObsReader.extensions,
       file_description=MagObsReader.description)
class PVGeoMagObsReader(MagObsReader):
    def __init__(self):
        MagObsReader.__init__(self)

    #### Seters and Geters ####
    @smproperty.xml(_helpers.get_file_reader_xml(MagObsReader.extensions, reader_description=MagObsReader.description))
    def AddFileName(self, filename):
        MagObsReader.AddFileName(self, filename)

    @smproperty.doublevector(name="TimeDelta", default_values=1.0, panel_visibility="advanced")
    def set_time_delta(self, dt):
        MagObsReader.set_time_delta(self, dt)

    @smproperty.doublevector(name="TimestepValues", information_only="1", si_class="vtkSITimeStepsProperty")
    def get_time_step_values(self):
        """This is critical for registering the timesteps"""
        return MagObsReader.get_time_step_values(self)

    @smproperty.intvector(name="SkipRows", default_values=0, panel_visibility="advanced")
    def set_skip_rows(self, skip):
        MagObsReader.set_skip_rows(self, skip)

    @smproperty.stringvector(name="Comments", default_values="!", panel_visibility="advanced")
    def set_comments(self, identifier):
        MagObsReader.set_comments(self, identifier)


###############################################################################


@smproxy.filter(name='PVGeoGeologyMapper', label=GeologyMapper.__displayname__)
@smhint.xml('''<ShowInMenu category="%s"/>
    <RepresentationType view="RenderView" type="Surface" />''' % MENU_CAT)
@smproperty.input(name="Input", port_index=0)
@smdomain.datatype(dataTypes=["vtkDataObject"], composite_data_supported=False)
class PVGeoGeologyMapper(GeologyMapper):
    def __init__(self):
        GeologyMapper.__init__(self)

    #### SETTERS AND GETTERS ####

    @smproperty.xml(_helpers.get_input_array_xml(nInputPorts=1, n_arrays=1))
    def SetInputArrayToProcess(self, idx, port, connection, field, name):
        return GeologyMapper.SetInputArrayToProcess(self, idx, port, connection, field, name)

    @smproperty.xml('''
        <StringVectorProperty
            panel_visibility="default"
            name="FileName"
            label="File Name"
            command="SetFileName"
            animateable="1"
            number_of_elements="1">
            <FileListDomain name="filename"/>
            <Documentation>This is the file contating the mapping definitions.</Documentation>
        </StringVectorProperty>''')
    def SetFileName(self, filename):
        GeologyMapper.SetFileName(self, filename)

    @smproperty.stringvector(name="Delimiter", default_values=",", panel_visibility="advanced")
    def set_delimiter(self, deli):
        GeologyMapper.set_delimiter(self, deli)


###############################################################################


if discretize_available:
    @smproxy.reader(name="PVGeoDiscretizeMeshReader",
           label='PVGeo: %s'%DiscretizeMeshReader.__displayname__,
           extensions=DiscretizeMeshReader.extensions,
           file_description=DiscretizeMeshReader.description)
    class PVGeoDiscretizeMeshReader(DiscretizeMeshReader):
        def __init__(self):
            DiscretizeMeshReader.__init__(self)

        #### Seters and Geters ####

        @smproperty.xml(_helpers.get_file_reader_xml(DiscretizeMeshReader.extensions, reader_description=DiscretizeMeshReader.description))
        def AddFileName(self, filename):
            DiscretizeMeshReader.AddFileName(self, filename)

        @smproperty.doublevector(name="TimeDelta", default_values=1.0, panel_visibility="advanced")
        def set_time_delta(self, dt):
            DiscretizeMeshReader.set_time_delta(self, dt)

        @smproperty.doublevector(name="TimestepValues", information_only="1", si_class="vtkSITimeStepsProperty")
        def get_time_step_values(self):
            """This is critical for registering the timesteps"""
            return DiscretizeMeshReader.get_time_step_values(self)
